/* Attachments for the article "Evaluating the normal distribution" */

/*------------------------------------------------------
The following  little C function Phi(x)
   will provide double precision values for
   the standard normal distribution Phi(x)
   with absolute error less than 8*10^(-16).
--------------------------------------------------------*/
    double Phi(double x)
    {long double s=x,t=0,b=x,q=x*x,i=1;
     while(s!=t) s=(t=s)+(b*=q/(i+=2));
     return .5+s*exp(-.5*q-.91893853320467274178L);}

/*--------------------------------------------------------
   And this slightly larger cPhi(x) will provide
   values for the complementary normal distribution.
   Its accuracy, 14-16 digits, is limited primarily by
   the accuracy of exp(-t) for large t.
--------------------------------------------------------*/
       double cPhi(double x)
       {long double R[9]={1.25331413731550025L, .421369229288054473L, .236652382913560671L,
                        .162377660896867462L, .123131963257932296L,.0990285964717319214L,
                        .0827662865013691773L,.0710695805388521071L,.0622586659950261958L};
       int i,j=.5*(fabs(x)+1);
       long double pwr=1,a=R[j],z=2*j,b=a*z-1,h=fabs(x)-z,s=a+h*b,t=a,q=h*h;
       for(i=2;s!=t;i+=2){a=(a+z*b)/i; b=(b+z*a)/(i+1); pwr*=q; s=(t=s)+pwr*(a+h*b);}
       s=s*exp(-.5*x*x-.91893853320467274178L);
       if(x>=0) return (double) s; return (double) (1.-s); }

/*--------------------------------------------------
 You will find those two little routines much simpler,
 with accuracy equal to, and often better, than methods
 based on erf or erfc functions. Several of the more common
 erf or erfc's  are listed below, some of them
 adapted to provide the standard normal distribution
 directly, without the need for .5+.5*erf(x/sqrt(2.))
 or .5*erfc(x/sqrt(2.)).
 Furthermore, Phi and cPhi are based on familiar
 calculus-level mathematics as developed in the
 attached article, avoiding the need for mysterious
 and often ponderous developments based on polynomial,
 rational function or continued fraction expansions
 for erf or erfc functions.  This is not to belittle
 the difficulty or skill needed for such methods; they
 are required for many of the classical functions.
 It just happens, as is the theme of the attached article,
 that the normal distribution can be evaluated as accurately
 with a few C instructions based on easily understood mathematics.
 -------------------------------------------------------*/

/*--------------------------------------------------------
 Below are listings for the erf or erfc (or normal distribution
 functions based on erf or erfc) mentioned in the article, with
 accuracies tabled in the article or plotted in Figures 2,3,4,5.
 --------------------------------------------------------*/

/*---------------------------------------------------------------------------
              BEGIN C PROC DESIGNATED "CODY" IN ARTICLE
    Note: This listing corrects the value "double sixten=1.6;" to the proper
    value "double sixten=16.;".  (If "sixten" had been properly spelled "sixteen",
    the error might have been more easily discovered.)
----------------------------------------------------------------------------*/

/*
 *  R : A Computer Langage for Statistical Data Analysis
 *  Copyright (C) 1995, 1996  Robert Gentleman and Ross Ihaka
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* Reference:
 * Cody, W.D. (1993). ALGORITHM 715: SPECFUN - A Portable FORTRAN
 * Package of Special Function Routines and Test Drivers"
 * ACM Transactions on Mathematical Software. 19, 22-32.
 *
 * This function evaluates the normal distribution function:
 * The main computation evaluates near-minimax approximations
 * derived from those in "Rational Chebyshev approximations for
 * the error function" by W. J. Cody, Math. Comp., 1969, 631-637.
 * This transportable program uses rational functions that
 * theoretically approximate the normal distribution function to
 * at least 18 significant decimal digits.  The accuracy achieved
 * depends on the arithmetic system, the compiler, the intrinsic
 * functions, and proper selection of the machine-dependent
 * constants.
 *
 *
 * Mathematical Constants:
 *
 * sqrpi = 1 / sqrt(2*pi),
 * root32 = sqrt(32),
 * thrsh = the argument for which pnorm(thrsh,0,1) = 0.75.
 */


/*float.h:#define DBL_EPSILON 2.2204460492503131e-16*/
/*float.h:#define DBL_MIN 2.2250738585072014e-308*/

/*#include "Mathlib.h"*/
#include <float.h>
#include <math.h>

static double c[9] = {
	0.39894151208813466764,
	8.8831497943883759412,
	93.506656132177855979,
	597.27027639480026226,
	2494.5375852903726711,
	6848.1904505362823326,
	11602.651437647350124,
	9842.7148383839780218,
	1.0765576773720192317e-8
};

static double d[8] = {
	22.266688044328115691,
	235.38790178262499861,
	1519.377599407554805,
	6485.558298266760755,
	18615.571640885098091,
	34900.952721145977266,
	38912.003286093271411,
	19685.429676859990727
};

static double p[6] = {
	0.21589853405795699,
	0.1274011611602473639,
	0.022235277870649807,
	0.001421619193227893466,
	2.9112874951168792e-5,
	0.02307344176494017303
};

static double q[5] = {
	1.28426009614491121,
	0.468238212480865118,
	0.0659881378689285515,
	0.00378239633202758244,
	7.29751555083966205e-5
};

static double a[5] = {
	2.2352520354606839287,
	161.02823106855587881,
	1067.6894854603709582,
	18154.981253343561249,
	0.065682337918207449113
};

static double b[4] = {
	47.20258190468824187,
	976.09855173777669322,
	10260.932208618978205,
	45507.789335026729956}
;

static double one = 1.0;
static double half = 0.5;
static double zero = 0.0;
static double sixten = 16.; /* note correct value for "sixteen", not the original 1.6./*
static double sqrpi = 0.39894228040143267794;
static double thrsh = 0.66291;
static double root32 = 5.656854248;

/*double pnorm(double x, double mean, double sd)*/
double pnorms(x)
double x;
{
	static double xden, temp, xnum, result, ccum;
	static double del, min, eps, xsq;
	static double y;
	static int i;
	double fint();

/*	if(sd <= 0.0)
		DOMAIN_ERROR;
	x = (x - mean) / sd;*/
	eps = DBL_EPSILON * .5;
	min = DBL_MIN;
	y = fabs(x);
	if (y <= thrsh) {
		/* Evaluate pnorm for |z| <= 0.66291 */
		xsq = zero;
		if (y > eps) {
			xsq = x * x;
		}
		xnum = a[4] * xsq;
		xden = xsq;
		for (i = 1; i <= 3; ++i) {
			xnum = (xnum + a[i - 1]) * xsq;
			xden = (xden + b[i - 1]) * xsq;
		}
		result = x * (xnum + a[3]) / (xden + b[3]);
		temp = result;
		result = half + temp;
		ccum = half - temp;
	}
	else if (y <= root32) {
		/* Evaluate pnorm for 0.66291 <= |z| <= sqrt(32) */
		xnum = c[8] * y;
		xden = y;
		for (i = 1; i <= 7; ++i) {
			xnum = (xnum + c[i - 1]) * y;
			xden = (xden + d[i - 1]) * y;
		}
		result = (xnum + c[7]) / (xden + d[7]);
		xsq = fint(y * sixten) / sixten;
		del = (y - xsq) * (y + xsq);
		result = exp(-xsq * xsq * half) * exp(-del * half) * result;
		ccum = one - result;
		if (x > zero) {
			temp = result;
			result = ccum;
			ccum = temp;
		}
	}
	else {
		/* Evaluate pnorm for |z| > sqrt(32) */
		result = zero;
		xsq = one / (x * x);
		xnum = p[5] * xsq;
		xden = xsq;
		for (i = 1; i <= 4; ++i) {
			xnum = (xnum + p[i - 1]) * xsq;
			xden = (xden + q[i - 1]) * xsq;
		}
		result = xsq * (xnum + p[4]) / (xden + q[4]);
		result = (sqrpi - result) / y;
		xsq = fint(x * sixten) / sixten;
		del = (x - xsq) * (x + xsq);
		result = exp(-xsq * xsq * half) * exp(-del * half) * result;
		ccum = one - result;
		if (x > zero) {
			temp = result;
			result = ccum;
			ccum = temp;
		}
	}
	if (result < min) {
		result = 0.0;
	}
	if (ccum < min) {
		ccum = 0.0;
	}
	return result;
}

/*double fint(double x)*/
double fint(x)
double x;
{
        return (x >= 0.0) ? floor(x) : -floor(-x);
}

/*---------------------------------------------------------------------------
             BEGIN FORTRAN PROC DESIGNATED "AS66" IN ARTICLE
------------------------------------------------------------------------------*\

      DOUBLE PRECISION FUNCTION DERFC(X)
C***BEGIN PROLOGUE  DERFC
C***DATE WRITTEN   770701   (YYMMDD)
C***REVISION DATE  820801   (YYMMDD)
C***CATEGORY NO.  C8A,L5A1E
C***KEYWORDS  COMPLEMENTARY ERROR FUNCTION,DOUBLE PRECISION,
C             SPECIAL FUNCTION
C***AUTHOR  FULLERTON, W., (LANL)
C***PURPOSE  Computes the d.p. complementary error function, ERFC.
C***DESCRIPTION
C
C DERFC(X) calculates the double precision complementary error function
C for double precision argument X.
C
C Series for ERF        on the interval  0.          to  1.00000E+00
C                                        with weighted Error   1.28E-32
C                                         log weighted Error  31.89
C                               significant figures required  31.05
C                                    decimal places required  32.55
C
C Series for ERC2       on the interval  2.50000E-01 to  1.00000E+00
C                                        with weighted Error   2.67E-32
C                                         log weighted Error  31.57
C                               significant figures required  30.31
C                                    decimal places required  32.42
C
C Series for ERFC       on the interval  0.          to  2.50000E-01
C                                        with weighted error   1.53E-31
C                                         log weighted error  30.82
C                               significant figures required  29.47
C                                    decimal places required  31.70
C***REFERENCES  (NONE)
C***ROUTINES CALLED  D1MACH,DCSEVL,INITDS,XERROR
C***END PROLOGUE  DERFC
      DOUBLE PRECISION X, ERFCS(21), ERFCCS(59), ERC2CS(49), SQEPS,
     1  SQRTPI, XMAX, XSML, Y,  D1MACH, DCSEVL
      DATA ERF CS(  1) / -.4904612123 4691808039 9845440333 76 D-1     /
      DATA ERF CS(  2) / -.1422612051 0371364237 8247418996 31 D+0     /
      DATA ERF CS(  3) / +.1003558218 7599795575 7546767129 33 D-1     /
      DATA ERF CS(  4) / -.5768764699 7674847650 8270255091 67 D-3     /
      DATA ERF CS(  5) / +.2741993125 2196061034 4221607914 71 D-4     /
      DATA ERF CS(  6) / -.1104317550 7344507604 1353812959 05 D-5     /
      DATA ERF CS(  7) / +.3848875542 0345036949 9613114981 74 D-7     /
      DATA ERF CS(  8) / -.1180858253 3875466969 6317518015 81 D-8     /
      DATA ERF CS(  9) / +.3233421582 6050909646 4029309533 54 D-10    /
      DATA ERF CS( 10) / -.7991015947 0045487581 6073747085 95 D-12    /
      DATA ERF CS( 11) / +.1799072511 3961455611 9672454866 34 D-13    /
      DATA ERF CS( 12) / -.3718635487 8186926382 3168282094 93 D-15    /
      DATA ERF CS( 13) / +.7103599003 7142529711 6899083946 66 D-17    /
      DATA ERF CS( 14) / -.1261245511 9155225832 4954248533 33 D-18    /
      DATA ERF CS( 15) / +.2091640694 1769294369 1705002666 66 D-20    /
      DATA ERF CS( 16) / -.3253973102 9314072982 3641600000 00 D-22    /
      DATA ERF CS( 17) / +.4766867209 7976748332 3733333333 33 D-24    /
      DATA ERF CS( 18) / -.6598012078 2851343155 1999999999 99 D-26    /
      DATA ERF CS( 19) / +.8655011469 9637626197 3333333333 33 D-28    /
      DATA ERF CS( 20) / -.1078892517 7498064213 3333333333 33 D-29    /
      DATA ERF CS( 21) / +.1281188399 3017002666 6666666666 66 D-31    /
      DATA ERC2CS(  1) / -.6960134660 2309501127 3915082619 7 D-1      /
      DATA ERC2CS(  2) / -.4110133936 2620893489 8221208466 6 D-1      /
      DATA ERC2CS(  3) / +.3914495866 6896268815 6114370524 4 D-2      /
      DATA ERC2CS(  4) / -.4906395650 5489791612 8093545077 4 D-3      /
      DATA ERC2CS(  5) / +.7157479001 3770363807 6089414182 5 D-4      /
      DATA ERC2CS(  6) / -.1153071634 1312328338 0823284791 2 D-4      /
      DATA ERC2CS(  7) / +.1994670590 2019976350 5231486770 9 D-5      /
      DATA ERC2CS(  8) / -.3642666471 5992228739 3611843071 1 D-6      /
      DATA ERC2CS(  9) / +.6944372610 0050125899 3127721463 3 D-7      /
      DATA ERC2CS( 10) / -.1371220902 1043660195 3460514121 0 D-7      /
      DATA ERC2CS( 11) / +.2788389661 0071371319 6386034808 7 D-8      /
      DATA ERC2CS( 12) / -.5814164724 3311615518 6479105031 6 D-9      /
      DATA ERC2CS( 13) / +.1238920491 7527531811 8016881795 0 D-9      /
      DATA ERC2CS( 14) / -.2690639145 3067434323 9042493788 9 D-10     /
      DATA ERC2CS( 15) / +.5942614350 8479109824 4470968384 0 D-11     /
      DATA ERC2CS( 16) / -.1332386735 7581195792 8775442057 0 D-11     /
      DATA ERC2CS( 17) / +.3028046806 1771320171 7369724330 4 D-12     /
      DATA ERC2CS( 18) / -.6966648814 9410325887 9586758895 4 D-13     /
      DATA ERC2CS( 19) / +.1620854541 0539229698 1289322762 8 D-13     /
      DATA ERC2CS( 20) / -.3809934465 2504919998 7691305772 9 D-14     /
      DATA ERC2CS( 21) / +.9040487815 9788311493 6897101297 5 D-15     /
      DATA ERC2CS( 22) / -.2164006195 0896073478 0981204700 3 D-15     /
      DATA ERC2CS( 23) / +.5222102233 9958549846 0798024417 2 D-16     /
      DATA ERC2CS( 24) / -.1269729602 3645553363 7241552778 0 D-16     /
      DATA ERC2CS( 25) / +.3109145504 2761975838 3622741295 1 D-17     /
      DATA ERC2CS( 26) / -.7663762920 3203855240 0956671481 1 D-18     /
      DATA ERC2CS( 27) / +.1900819251 3627452025 3692973329 0 D-18     /
      DATA ERC2CS( 28) / -.4742207279 0690395452 2565599996 5 D-19     /
      DATA ERC2CS( 29) / +.1189649200 0765283828 8068307845 1 D-19     /
      DATA ERC2CS( 30) / -.3000035590 3257802568 4527131306 6 D-20     /
      DATA ERC2CS( 31) / +.7602993453 0432461730 1938527709 8 D-21     /
      DATA ERC2CS( 32) / -.1935909447 6068728815 6981104913 0 D-21     /
      DATA ERC2CS( 33) / +.4951399124 7733378810 0004238677 3 D-22     /
      DATA ERC2CS( 34) / -.1271807481 3363718796 0862198988 8 D-22     /
      DATA ERC2CS( 35) / +.3280049600 4695130433 1584165205 3 D-23     /
      DATA ERC2CS( 36) / -.8492320176 8228965689 2479242239 9 D-24     /
      DATA ERC2CS( 37) / +.2206917892 8075602235 1987998719 9 D-24     /
      DATA ERC2CS( 38) / -.5755617245 6965284983 1281950719 9 D-25     /
      DATA ERC2CS( 39) / +.1506191533 6392342503 5414405119 9 D-25     /
      DATA ERC2CS( 40) / -.3954502959 0187969531 0428569599 9 D-26     /
      DATA ERC2CS( 41) / +.1041529704 1515009799 8464505173 3 D-26     /
      DATA ERC2CS( 42) / -.2751487795 2787650794 5017890133 3 D-27     /
      DATA ERC2CS( 43) / +.7290058205 4975574089 9770368000 0 D-28     /
      DATA ERC2CS( 44) / -.1936939645 9159478040 7750109866 6 D-28     /
      DATA ERC2CS( 45) / +.5160357112 0514872983 7005482666 6 D-29     /
      DATA ERC2CS( 46) / -.1378419322 1930940993 8964480000 0 D-29     /
      DATA ERC2CS( 47) / +.3691326793 1070690422 5109333333 3 D-30     /
      DATA ERC2CS( 48) / -.9909389590 6243654206 5322666666 6 D-31     /
      DATA ERC2CS( 49) / +.2666491705 1953884133 2394666666 6 D-31     /
      DATA ERFCCS(  1) / +.7151793102 0292477450 3697709496 D-1        /
      DATA ERFCCS(  2) / -.2653243433 7606715755 8893386681 D-1        /
      DATA ERFCCS(  3) / +.1711153977 9208558833 2699194606 D-2        /
      DATA ERFCCS(  4) / -.1637516634 5851788416 3746404749 D-3        /
      DATA ERFCCS(  5) / +.1987129350 0552036499 5974806758 D-4        /
      DATA ERFCCS(  6) / -.2843712412 7665550875 0175183152 D-5        /
      DATA ERFCCS(  7) / +.4606161308 9631303696 9379968464 D-6        /
      DATA ERFCCS(  8) / -.8227753025 8792084205 7766536366 D-7        /
      DATA ERFCCS(  9) / +.1592141872 7709011298 9358340826 D-7        /
      DATA ERFCCS( 10) / -.3295071362 2528432148 6631665072 D-8        /
      DATA ERFCCS( 11) / +.7223439760 4005554658 1261153890 D-9        /
      DATA ERFCCS( 12) / -.1664855813 3987295934 4695966886 D-9        /
      DATA ERFCCS( 13) / +.4010392588 2376648207 7671768814 D-10       /
      DATA ERFCCS( 14) / -.1004816214 4257311327 2170176283 D-10       /
      DATA ERFCCS( 15) / +.2608275913 3003338085 9341009439 D-11       /
      DATA ERFCCS( 16) / -.6991110560 4040248655 7697812476 D-12       /
      DATA ERFCCS( 17) / +.1929492333 2617070862 4205749803 D-12       /
      DATA ERFCCS( 18) / -.5470131188 7543310649 0125085271 D-13       /
      DATA ERFCCS( 19) / +.1589663309 7626974483 9084032762 D-13       /
      DATA ERFCCS( 20) / -.4726893980 1975548392 0369584290 D-14       /
      DATA ERFCCS( 21) / +.1435873376 7849847867 2873997840 D-14       /
      DATA ERFCCS( 22) / -.4449510561 8173583941 7250062829 D-15       /
      DATA ERFCCS( 23) / +.1404810884 7682334373 7305537466 D-15       /
      DATA ERFCCS( 24) / -.4513818387 7642108962 5963281623 D-16       /
      DATA ERFCCS( 25) / +.1474521541 0451330778 7018713262 D-16       /
      DATA ERFCCS( 26) / -.4892621406 9457761543 6841552532 D-17       /
      DATA ERFCCS( 27) / +.1647612141 4106467389 5301522827 D-17       /
      DATA ERFCCS( 28) / -.5626817176 3294080929 9928521323 D-18       /
      DATA ERFCCS( 29) / +.1947443382 2320785142 9197867821 D-18       /
      DATA ERFCCS( 30) / -.6826305642 9484207295 6664144723 D-19       /
      DATA ERFCCS( 31) / +.2421988887 2986492401 8301125438 D-19       /
      DATA ERFCCS( 32) / -.8693414133 5030704256 3800861857 D-20       /
      DATA ERFCCS( 33) / +.3155180346 2280855712 2363401262 D-20       /
      DATA ERFCCS( 34) / -.1157372324 0496087426 1239486742 D-20       /
      DATA ERFCCS( 35) / +.4288947161 6056539462 3737097442 D-21       /
      DATA ERFCCS( 36) / -.1605030742 0576168500 5737770964 D-21       /
      DATA ERFCCS( 37) / +.6063298757 4538026449 5069923027 D-22       /
      DATA ERFCCS( 38) / -.2311404251 6979584909 8840801367 D-22       /
      DATA ERFCCS( 39) / +.8888778540 6618855255 4702955697 D-23       /
      DATA ERFCCS( 40) / -.3447260576 6513765223 0718495566 D-23       /
      DATA ERFCCS( 41) / +.1347865460 2069650682 7582774181 D-23       /
      DATA ERFCCS( 42) / -.5311794071 1250217364 5873201807 D-24       /
      DATA ERFCCS( 43) / +.2109341058 6197831682 8954734537 D-24       /
      DATA ERFCCS( 44) / -.8438365587 9237891159 8133256738 D-25       /
      DATA ERFCCS( 45) / +.3399982524 9452089062 7359576337 D-25       /
      DATA ERFCCS( 46) / -.1379452388 0732420900 2238377110 D-25       /
      DATA ERFCCS( 47) / +.5634490311 8332526151 3392634811 D-26       /
      DATA ERFCCS( 48) / -.2316490434 4770654482 3427752700 D-26       /
      DATA ERFCCS( 49) / +.9584462844 6018101526 3158381226 D-27       /
      DATA ERFCCS( 50) / -.3990722880 3301097262 4224850193 D-27       /
      DATA ERFCCS( 51) / +.1672129225 9444773601 7228709669 D-27       /
      DATA ERFCCS( 52) / -.7045991522 7660138563 8803782587 D-28       /
      DATA ERFCCS( 53) / +.2979768402 8642063541 2357989444 D-28       /
      DATA ERFCCS( 54) / -.1262522466 4606192972 2422632994 D-28       /
      DATA ERFCCS( 55) / +.5395438704 5424879398 5299653154 D-29       /
      DATA ERFCCS( 56) / -.2380992882 5314591867 5346190062 D-29       /
      DATA ERFCCS( 57) / +.1099052830 1027615735 9726683750 D-29       /
      DATA ERFCCS( 58) / -.4867713741 6449657273 2518677435 D-30       /
      DATA ERFCCS( 59) / +.1525877264 1103575676 3200828211 D-30       /
      DATA SQRTPI / 1.772453850 9055160272 9816748334 115D0 /
      DATA NTERF, NTERFC, NTERC2, XSML, XMAX, SQEPS / 3*0, 3*0.D0 /
C***FIRST EXECUTABLE STATEMENT  DERFC
      IF (NTERF.NE.0) GO TO 10
      ETA = 0.1*SNGL(D1MACH(3))
      NTERF = INITDS (ERFCS, 21, ETA)
      NTERFC = INITDS (ERFCCS, 59, ETA)
      NTERC2 = INITDS (ERC2CS, 49, ETA)
C
      XSML = -DSQRT (-DLOG(SQRTPI*D1MACH(3)))
      XMAX = DSQRT (-DLOG(SQRTPI*D1MACH(1)) )
      XMAX = XMAX - 0.5D0*DLOG(XMAX)/XMAX - 0.01D0
      SQEPS = DSQRT (2.0D0*D1MACH(3))
C
 10   IF (X.GT.XSML) GO TO 20
C
C ERFC(X) = 1.0 - ERF(X)  FOR  X .LT. XSML
C
      DERFC = 2.0D0
      RETURN
C
 20   IF (X.GT.XMAX) GO TO 40
      Y = DABS(X)
      IF (Y.GT.1.0D0) GO TO 30
C
C ERFC(X) = 1.0 - ERF(X)  FOR ABS(X) .LE. 1.0
C
      IF (Y.LT.SQEPS) DERFC = 1.0D0 - 2.0D0*X/SQRTPI
      IF (Y.GE.SQEPS) DERFC = 1.0D0 - X*(1.0D0 + DCSEVL (2.D0*X*X-1.D0,
     1  ERFCS, NTERF))
      RETURN
C
C ERFC(X) = 1.0 - ERF(X)  FOR  1.0 .LT. ABS(X) .LE. XMAX
C
 30   Y = Y*Y
      IF (Y.LE.4.D0) DERFC = DEXP(-Y)/DABS(X) * (0.5D0 + DCSEVL (
     1  (8.D0/Y-5.D0)/3.D0, ERC2CS, NTERC2) )
      IF (Y.GT.4.D0) DERFC = DEXP(-Y)/DABS(X) * (0.5D0 + DCSEVL (
     1  8.D0/Y-1.D0, ERFCCS, NTERFC) )
      IF (X.LT.0.D0) DERFC = 2.0D0 - DERFC
      RETURN
C
 40   CALL XERROR ( 'DERFC   X SO BIG ERFC UNDERFLOWS', 32, 1, 1)
      DERFC = 0.D0
      RETURN
C
      END



/*---------------------------------------------------------------------------
              BEGIN C PROC DESIGNATED "HART" IN ARTICLE
----------------------------------------------------------------------------*/

#define SPLIT (double)8.0

/*The following coefficients are from COMPUTER APPROXIMATIONS, */
/* by J.F. Hart and others, Wiley, 1968 (QA297.C64) */
/* ERFC 5662, p.292, AND ERFC 5725, p.297. */

double Plo[] = {
1.826334884229511E+03 , 2.898029329216766E+03 , 2.320439590251635E+03
,
1.143262070703886E+03 , 3.685196154710010E+02 , 7.708161730368429E+01
,
9.675807882987265E+00 , 5.641877825507398E-01
};
int PloOrd = 7;

double Qlo[] = {
1.826334884229511E+03 , 4.958827564721141E+03 , 6.089542423272444E+03
,
4.429612803883683E+03 , 2.094384367789540E+03 , 6.617361207107654E+02
,
1.371255960500622E+02 , 1.714980943627608E+01 , 1.000000000000000E+00
};
int QloOrd = 8;

double Phi[] = {
2.978865626393993E+00 , 7.409740605964742E+00 , 6.160209853109631E+00
,
5.019049726784267E+00 , 1.275366644729966E+00 , 5.641895835477551E-01
};
int PhiOrd = 5;

double Qhi[] = {
3.369075206982753E+00 , 9.608965327192788E+00 , 1.708144074746600E+01
,
1.204895192785513E+01 , 9.396034016235054E+00 , 2.260528520767327E+00
,
1.000000000000000E+00
};
int QhiOrd = 6;

/* ===============================================================
/*Test main program for Normal16 against ConSA Standard full */
/* precision function mdnor(x). */
main() {
  double x, ans, ans2, diff, Normal16(double x), mdnor(double x);
  while (1) {
      scanf("%lf", &x);
      ans = Normal16(x);
      ans2 = 0.; //mdnor(x);
      diff = ans - ans2;
      printf("%18.16f  %8.2e  %5.2f\n", ans, diff, x);
  }
}

/*=============================================================== */

double Normal16(double x) {
  double arg, ans, Erf16(double x);
  arg = x / sqrt(2.0);
  if (x < 0.) {
      x = -x;
      arg = -arg;
      ans = .5 * Erf16(arg);
  }
  else
      ans = 1.0 - .5 * Erf16(arg);

  return(ans);
}

double Erf16(double x) {
  double ans, Eval(double *A, int N, double x);

  ans = exp(-x*x);

  if (x >= SPLIT)
      ans *= (Eval(Phi, PhiOrd, x) / Eval(Qhi, QhiOrd, x) );
  else
      ans *= (Eval(Plo, PloOrd, x) / Eval(Qlo, QloOrd, x) );

  return (ans);
}

double Eval(double *A, int N, double x) {
  double ans;
  int    i;
  ans = 0.;
  for (i=N; i>0; i--)
      ans = (ans + A[i]) * x;
  ans += A[0];
  return(ans);
}


/*-----------------------------------------------------------------------
               BEGIN  C PROC DESIGNATED "DERFC" IN ARTICLE
    I found this the best of the lot, almost as short as cPhi
    and as accurate.  The mysterious rational function of
    t = 3.97886080735226/(fabs(x) + 3.97886080735226) can be
    admired but not easily understood.
------------------------------------------------------------------------*/
/* error function in double precision */
//Copyright(C) 1996 Takuya OOURA (email: ooura@mmm.t.u-tokyo.ac.jp).
// from http://momonga.t.u-tokyo.ac.jp/~ooura/gamerf.html

double derfc(double x)
{
    double t, u, y;

    t = 3.97886080735226/(fabs(x) + 3.97886080735226);
    u = t - 0.5;
    y = (((((((((0.00127109764952614092*u + 1.19314022838340944e-4)*u -
        0.003963850973605135)*u - 8.70779635317295828e-4)*u +
        0.00773672528313526668)*u + 0.00383335126264887303)*u -
        0.0127223813782122755)*u - 0.0133823644533460069)*u +
        0.0161315329733252248)*u + 0.0390976845588484035)*u +
        0.00249367200053503304;
    y = ((((((((((((y*u - 0.0838864557023001992)*u -
        0.119463959964325415)*u + 0.0166207924969367356)*u +
        0.357524274449531043)*u + 0.805276408752910567)*u +
        1.18902982909273333)*u + 1.37040217682338167)*u +
        1.31314653831023098)*u + 1.07925515155856677)*u +
        0.774368199119538609)*u + 0.490165080585318424)*u +
        0.275374741597376782)*t*exp(-x*x);
    return x < 0 ? 2-y : y;
}

/*---------------------------------------------------------------------------
            BEGIN C PROCS ERF AND ERFC DESIGNATED "SUN" IN ARTICLE
----------------------------------------------------------------------------*/

/* ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/* double erf(double x)
 * double erfc(double x)
 *        x
 *        2      |\
 *     erf(x)  =  ---------  | exp(-t*t)dt
 *      sqrt(pi) \|
 *        0
 *
 *     erfc(x) =  1-erf(x)
 *  Note that
 *  erf(-x) = -erf(x)
 *  erfc(-x) = 2 - erfc(x)
 *
 * Method:
 * 1. For |x| in [0, 0.84375]
 *     erf(x)  = x + x*R(x^2)
 *          erfc(x) = 1 - erf(x)           if x in [-.84375,0.25]
 *                  = 0.5 + ((0.5-x)-x*R)  if x in [0.25,0.84375]
 *    where R = P/Q where P is an odd poly of degree 8 and
 *    Q is an odd poly of degree 10.
 *       -57.90
 *   | R - (erf(x)-x)/x | <= 2
 *
 *
 *    Remark. The formula is derived by noting
 *          erf(x) = (2/sqrt(pi))*(x - x^3/3 + x^5/10 - x^7/42 + ....)
 *    and that
 *          2/sqrt(pi) = 1.128379167095512573896158903121545171688
 *    is close to one. The interval is chosen because the fix
 *    point of erf(x) is near 0.6174 (i.e., erf(x)=x when x is
 *    near 0.6174), and by some experiment, 0.84375 is chosen to
 *     guarantee the error is less than one ulp for erf.
 *
 *      2. For |x| in [0.84375,1.25], let s = |x| - 1, and
 *         c = 0.84506291151 rounded to single (24 bits)
 *          erf(x)  = sign(x) * (c  + P1(s)/Q1(s))
 *          erfc(x) = (1-c)  - P1(s)/Q1(s) if x > 0
 *     1+(c+P1(s)/Q1(s))    if x < 0
 *          |P1/Q1 - (erf(|x|)-c)| <= 2**-59.06
 *    Remark: here we use the taylor series expansion at x=1.
 *  erf(1+s) = erf(1) + s*Poly(s)
 *    = 0.845.. + P1(s)/Q1(s)
 *    That is, we use rational approximation to approximate
 *   erf(1+s) - (c = (single)0.84506291151)
 *    Note that |P1/Q1|< 0.078 for x in [0.84375,1.25]
 *    where
 *  P1(s) = degree 6 poly in s
 *  Q1(s) = degree 6 poly in s
 *
 *      3. For x in [1.25,1/0.35(~2.857143)],
 *          erfc(x) = (1/x)*exp(-x*x-0.5625+R1/S1)
 *          erf(x)  = 1 - erfc(x)
 *    where
 *  R1(z) = degree 7 poly in z, (z=1/x^2)
 *  S1(z) = degree 8 poly in z
 *
 *      4. For x in [1/0.35,28]
 *          erfc(x) = (1/x)*exp(-x*x-0.5625+R2/S2) if x > 0
 *   = 2.0 - (1/x)*exp(-x*x-0.5625+R2/S2) if -6<x<0
 *   = 2.0 - tiny  (if x <= -6)
 *          erf(x)  = sign(x)*(1.0 - erfc(x)) if x < 6, else
 *          erf(x)  = sign(x)*(1.0 - tiny)
 *    where
 *  R2(z) = degree 6 poly in z, (z=1/x^2)
 *  S2(z) = degree 7 poly in z
 *
 *      Note1:
 *    To compute exp(-x*x-0.5625+R/S), let s be a single
 *    precision number and s := x; then
 *  -x*x = -s*s + (s-x)*(s+x)
 *         exp(-x*x-0.5626+R/S) =
 *   exp(-s*s-0.5625)*exp((s-x)*(s+x)+R/S);
 *      Note2:
 *    Here 4 and 5 make use of the asymptotic series
 *     exp(-x*x)
 *  erfc(x) ~ ---------- * ( 1 + Poly(1/x^2) )
 *     x*sqrt(pi)
 *    We use rational approximation to approximate
 *       g(s)=f(1/x^2) = log(erfc(x)*x) - x*x + 0.5625
 *    Here is the error bound for R1/S1 and R2/S2
 *       |R1/S1 - f(x)|  < 2**(-62.57)
 *       |R2/S2 - f(x)|  < 2**(-61.52)
 *
 *      5. For inf > x >= 28
 *          erf(x)  = sign(x) *(1 - tiny)  (raise inexact)
 *          erfc(x) = tiny*tiny (raise underflow) if x > 0
 *   = 2 - tiny if x<0
 *
 *      7. Special case:
 *          erf(0)  = 0, erf(inf)  = 1, erf(-inf) = -1,
 *          erfc(0) = 1, erfc(inf) = 0, erfc(-inf) = 2,
 *     erfc/erf(NaN) is NaN
 */


static const double tiny     = 1e-300,
half=  5.00000000000000000000e-01, /* 0x3FE00000, 0x00000000 */
one =  1.00000000000000000000e+00, /* 0x3FF00000, 0x00000000 */
two =  2.00000000000000000000e+00, /* 0x40000000, 0x00000000 */
 /* c = (float)0.84506291151 */
erx =  8.45062911510467529297e-01, /* 0x3FEB0AC1, 0x60000000 */
/*
 * Coefficients for approximation to  erf on [0,0.84375]
 */
efx =  1.28379167095512586316e-01, /* 0x3FC06EBA, 0x8214DB69 */
efx8=  1.02703333676410069053e+00, /* 0x3FF06EBA, 0x8214DB69 */
pp0  =  1.28379167095512558561e-01, /* 0x3FC06EBA, 0x8214DB68 */
pp1  = -3.25042107247001499370e-01, /* 0xBFD4CD7D, 0x691CB913 */
pp2  = -2.84817495755985104766e-02, /* 0xBF9D2A51, 0xDBD7194F */
pp3  = -5.77027029648944159157e-03, /* 0xBF77A291, 0x236668E4 */
pp4  = -2.37630166566501626084e-05, /* 0xBEF8EAD6, 0x120016AC */
qq1  =  3.97917223959155352819e-01, /* 0x3FD97779, 0xCDDADC09 */
qq2  =  6.50222499887672944485e-02, /* 0x3FB0A54C, 0x5536CEBA */
qq3  =  5.08130628187576562776e-03, /* 0x3F74D022, 0xC4D36B0F */
qq4  =  1.32494738004321644526e-04, /* 0x3F215DC9, 0x221C1A10 */
qq5  = -3.96022827877536812320e-06, /* 0xBED09C43, 0x42A26120 */
/*
 * Coefficients for approximation to  erf  in [0.84375,1.25]
 */
pa0  = -2.36211856075265944077e-03, /* 0xBF6359B8, 0xBEF77538 */
pa1  =  4.14856118683748331666e-01, /* 0x3FDA8D00, 0xAD92B34D */
pa2  = -3.72207876035701323847e-01, /* 0xBFD7D240, 0xFBB8C3F1 */
pa3  =  3.18346619901161753674e-01, /* 0x3FD45FCA, 0x805120E4 */
pa4  = -1.10894694282396677476e-01, /* 0xBFBC6398, 0x3D3E28EC */
pa5  =  3.54783043256182359371e-02, /* 0x3FA22A36, 0x599795EB */
pa6  = -2.16637559486879084300e-03, /* 0xBF61BF38, 0x0A96073F */
qa1  =  1.06420880400844228286e-01, /* 0x3FBB3E66, 0x18EEE323 */
qa2  =  5.40397917702171048937e-01, /* 0x3FE14AF0, 0x92EB6F33 */
qa3  =  7.18286544141962662868e-02, /* 0x3FB2635C, 0xD99FE9A7 */
qa4  =  1.26171219808761642112e-01, /* 0x3FC02660, 0xE763351F */
qa5  =  1.36370839120290507362e-02, /* 0x3F8BEDC2, 0x6B51DD1C */
qa6  =  1.19844998467991074170e-02, /* 0x3F888B54, 0x5735151D */
/*
 * Coefficients for approximation to  erfc in [1.25,1/0.35]
 */
ra0  = -9.86494403484714822705e-03, /* 0xBF843412, 0x600D6435 */
ra1  = -6.93858572707181764372e-01, /* 0xBFE63416, 0xE4BA7360 */
ra2  = -1.05586262253232909814e+01, /* 0xC0251E04, 0x41B0E726 */
ra3  = -6.23753324503260060396e+01, /* 0xC04F300A, 0xE4CBA38D */
ra4  = -1.62396669462573470355e+02, /* 0xC0644CB1, 0x84282266 */
ra5  = -1.84605092906711035994e+02, /* 0xC067135C, 0xEBCCABB2 */
ra6  = -8.12874355063065934246e+01, /* 0xC0545265, 0x57E4D2F2 */
ra7  = -9.81432934416914548592e+00, /* 0xC023A0EF, 0xC69AC25C */
sa1  =  1.96512716674392571292e+01, /* 0x4033A6B9, 0xBD707687 */
sa2  =  1.37657754143519042600e+02, /* 0x4061350C, 0x526AE721 */
sa3  =  4.34565877475229228821e+02, /* 0x407B290D, 0xD58A1A71 */
sa4  =  6.45387271733267880336e+02, /* 0x40842B19, 0x21EC2868 */
sa5  =  4.29008140027567833386e+02, /* 0x407AD021, 0x57700314 */
sa6  =  1.08635005541779435134e+02, /* 0x405B28A3, 0xEE48AE2C */
sa7  =  6.57024977031928170135e+00, /* 0x401A47EF, 0x8E484A93 */
sa8  = -6.04244152148580987438e-02, /* 0xBFAEEFF2, 0xEE749A62 */
/*
 * Coefficients for approximation to  erfc in [1/.35,28]
 */
rb0  = -9.86494292470009928597e-03, /* 0xBF843412, 0x39E86F4A */
rb1  = -7.99283237680523006574e-01, /* 0xBFE993BA, 0x70C285DE */
rb2  = -1.77579549177547519889e+01, /* 0xC031C209, 0x555F995A */
rb3  = -1.60636384855821916062e+02, /* 0xC064145D, 0x43C5ED98 */
rb4  = -6.37566443368389627722e+02, /* 0xC083EC88, 0x1375F228 */
rb5  = -1.02509513161107724954e+03, /* 0xC0900461, 0x6A2E5992 */
rb6  = -4.83519191608651397019e+02, /* 0xC07E384E, 0x9BDC383F */
sb1  =  3.03380607434824582924e+01, /* 0x403E568B, 0x261D5190 */
sb2  =  3.25792512996573918826e+02, /* 0x40745CAE, 0x221B9F0A */
sb3  =  1.53672958608443695994e+03, /* 0x409802EB, 0x189D5118 */
sb4  =  3.19985821950859553908e+03, /* 0x40A8FFB7, 0x688C246A */
sb5  =  2.55305040643316442583e+03, /* 0x40A3F219, 0xCEDF3BE6 */
sb6  =  4.74528541206955367215e+02, /* 0x407DA874, 0xE79FE763 */
sb7  = -2.24409524465858183362e+01; /* 0xC03670E2, 0x42712D62 */

extern double exp(double);
extern double fabs(double);
double erf(double x)
{
 int n0,hx,ix,i;
 double R,S,P,Q,s,y,z,r;
 n0 = ((*(int*)&one)>>29)^1;
 hx = *(n0+(int*)&x);
 ix = hx&0x7fffffff;
 if(ix>=0x7ff00000) {  /* erf(nan)=nan */
     i = ((unsigned)hx>>31)<<1;
     return (double)(1-i)+one/x; /* erf(+-inf)=+-1 */
 }

 if(ix < 0x3feb0000) {  /* |x|<0.84375 */
     if(ix < 0x3e300000) {  /* |x|<2**-28 */
         if (ix < 0x00800000)
      return 0.125*(8.0*x+efx8*x);  /*avoid underflow */
  return x + efx*x;
     }
     z = x*x;
     r = pp0+z*(pp1+z*(pp2+z*(pp3+z*pp4)));
     s = one+z*(qq1+z*(qq2+z*(qq3+z*(qq4+z*qq5))));
     y = r/s;
     return x + x*y;
 }
 if(ix < 0x3ff40000) {  /* 0.84375 <= |x| < 1.25 */
     s = fabs(x)-one;
     P = pa0+s*(pa1+s*(pa2+s*(pa3+s*(pa4+s*(pa5+s*pa6)))));
     Q = one+s*(qa1+s*(qa2+s*(qa3+s*(qa4+s*(qa5+s*qa6)))));
     if(hx>=0) return erx + P/Q; else return -erx - P/Q;
 }
 if (ix >= 0x40180000) {  /* inf>|x|>=6 */
     if(hx>=0) return one-tiny; else return tiny-one;
 }
 x = fabs(x);
  s = one/(x*x);
 if(ix< 0x4006DB6E) { /* |x| < 1/0.35 */
     R=ra0+s*(ra1+s*(ra2+s*(ra3+s*(ra4+s*(
    ra5+s*(ra6+s*ra7))))));
     S=one+s*(sa1+s*(sa2+s*(sa3+s*(sa4+s*(
    sa5+s*(sa6+s*(sa7+s*sa8)))))));
 } else { /* |x| >= 1/0.35 */
     R=rb0+s*(rb1+s*(rb2+s*(rb3+s*(rb4+s*(
    rb5+s*rb6)))));
     S=one+s*(sb1+s*(sb2+s*(sb3+s*(sb4+s*(
    sb5+s*(sb6+s*sb7))))));
 }
 z  = x;
 *(1-n0+(int*)&z) = 0;
 r  =  exp(-z*z-0.5625)*exp((z-x)*(z+x)+R/S);
 if(hx>=0) return one-r/x; else return  r/x-one;
}

double erfc(double x)
{
 int n0,hx,ix;
 double R,S,P,Q,s,y,z,r;
 n0 = ((*(int*)&one)>>29)^1;
 hx = *(n0+(int*)&x);
 ix = hx&0x7fffffff;
 if(ix>=0x7ff00000) {   /* erfc(nan)=nan */
      /* erfc(+-inf)=0,2 */
     return (double)(((unsigned)hx>>31)<<1)+one/x;
 }

 if(ix < 0x3feb0000) {  /* |x|<0.84375 */
     if(ix < 0x3c700000)   /* |x|<2**-56 */
  return one-x;
     z = x*x;
     r = pp0+z*(pp1+z*(pp2+z*(pp3+z*pp4)));
     s = one+z*(qq1+z*(qq2+z*(qq3+z*(qq4+z*qq5))));
     y = r/s;
     if(hx < 0x3fd00000) {   /* x<1/4 */
  return one-(x+x*y);
     } else {
  r = x*y;
  r += (x-half);
         return half - r ;
     }
 }
 if(ix < 0x3ff40000) {  /* 0.84375 <= |x| < 1.25 */
     s = fabs(x)-one;
     P = pa0+s*(pa1+s*(pa2+s*(pa3+s*(pa4+s*(pa5+s*pa6)))));
     Q = one+s*(qa1+s*(qa2+s*(qa3+s*(qa4+s*(qa5+s*qa6)))));
     if(hx>=0) {
         z  = one-erx; return z - P/Q;
     } else {
  z = erx+P/Q; return one+z;
     }
 }
 if (ix < 0x403c0000) {  /* |x|<28 */
     x = fabs(x);
      s = one/(x*x);
     if(ix< 0x4006DB6D) { /* |x| < 1/.35 ~ 2.857143*/
         R=ra0+s*(ra1+s*(ra2+s*(ra3+s*(ra4+s*(
    ra5+s*(ra6+s*ra7))))));
         S=one+s*(sa1+s*(sa2+s*(sa3+s*(sa4+s*(
    sa5+s*(sa6+s*(sa7+s*sa8)))))));
     } else {   /* |x| >= 1/.35 ~ 2.857143 */
  if(hx<0&&ix>=0x40180000) return two-tiny;/* x < -6 */
         R=rb0+s*(rb1+s*(rb2+s*(rb3+s*(rb4+s*(
    rb5+s*rb6)))));
         S=one+s*(sb1+s*(sb2+s*(sb3+s*(sb4+s*(
    sb5+s*(sb6+s*sb7))))));
     }
     z  = x;
     *(1-n0+(int*)&z)  = 0;
     r  =  exp(-z*z-0.5625)*
   exp((z-x)*(z+x)+R/S);
     if(hx>0) return r/x; else return two-r/x;
 } else {  if(hx>0) return tiny*tiny; else return two-tiny;
 }
}

